Documentation - Jobs
====================

Every job must implement the interface `SlmQueue\Job\JobInterface`. For ease of use, an abstract class 
`SlmQueue\Job\AbstractJob` is provided to implement most of the interface methods. The method, `execute()`, must be
implemented in userland code.

An example job would look like the following:

```php
namespace MyModule\Job;

use SlmQueue\Job\AbstractJob;

class PrintHelloWorldJob extends AbstractJob
{
    public function execute()
    {
        echo 'Hello World';
    }
}
```

Job payload
-----------
The job often needs some data to work with, commonly called a payload. For an email job, the contents of the email (to
address, subject and message) are the payload. If you generate an invoice based on an order in your database, you set
the id of this order as the payload. 

Internally, a `SlmQueue\Job\AbstractJob` extends the `Zend\Stdlib\Message` class which allows it to set and get content.
These methods are used to set the payload upon pushing and retrieve the payload when the job is executed.

To set the payload in a controller, below is an example:

```php
namespace MyModule\Controller;

use Zend\Mvc\Controller\AbstractActionController;

class MyController extends AbstractActionController
{
    public function fooAction()
    {
        // Do some work

        $job = $this->jobManager->get('MyModule\Job\SendEmailJob');
        $job->setContent(array(
            'to'      => 'bob@acme.com',
            'subject' => 'Registration completed',
            'message' => 'Hi bob, you just registered for our website! Welcome!'
         ));

        $this->queue->push($job);
    }
}
```

Then you can fetch this payload from the job itself:

```php
namespace MyModule\Job;

use SlmQueue\Job\AbstractJob;

class SendEmailJob extends AbstractJob
{
    public function execute()
    {
        $payload = $this->getContent();
        $to      = $payload['to'];
        $subject = $payload['subject'];
        $message = $payload['message'];

        // Do something with $to, $subjet, $message
    }
}
```

Job payload are a flexible array structure and will be automatically serialized by SlmQueue. This means you can have
values like a `DateTime` object as payload for jobs which will be serialized and deserialized in the background. However,
this also give you restrictions than payload data *must* be serializable. Doctrine 2 entities, references to 
`Zend\Mvc\Application` or other unserializable instances should not be set as payload.

Job dependencies
----------------

Because of the configuration via a job plugin manager, you can inject dependencies you need into the constructor of the
job class. This will require you to define a job factory for that job as well. Assume here we send an email in a job and
this job requires the email transport class as a dependency.

```php
namespace MyModule\Job;

use SlmQueue\Job\AbstractJob;
use Zend\Mail\Transport\TransportInterface;
use Zend\Mail\Message;

class SendEmailJob extends AbstractJob
{
    protected $transport;

    public function __construct(TransportInterface $transport)
    {
        $this->transport = $transport;
    }

    public function execute()
    {
        $message = new Message;
        $payload = $this->getContent();

        $message->setTo($payload['to']);
        $message->setSubject($payload['subject']);
        $message->setContent($payload['content']);

        $this->transport->send($message);
    }
}
```

To inject the email transport instance, a factory must be created to instantiate the job:

```php
namespace MyModule\Factory;

use Zend\ServiceManager\FactoryInterface;
use Zend\ServiceManager\ServiceLocatorInterface;

use MyModule\Job\SendEmailJob;

class SendEmailJobFactory implements FactoryInterface
{
    public function createService(ServiceLocatorInterface $sl)
    {
        $transport = $sl->getServiceLocator()->get('my-transport-service');

        $job = new SendEmailJob($transport);
        return $job;
    }
}
```

The last step is to register this factory for the above job. Note that in order for this to work the service should be registered with the FQCN of the job class as its identifier.

```php
'slm_queue' => array(
    'job_manager' => array(
        'factories' => array(
            'MyModule\Job\SendEmailJob' => 'MyModule\Factory\SendEmailJobFactory',
        ),
    ),
)
```

Push jobs via the controller plugin
-----------------------------------

Above examples all showed how to inject jobs using the queue directly. This requires SlmQueue users to inject the queue,
or queue plugin manager, in the controller. To avoid duplicate code about queue injection and job retrieval, a queue controller
plugin is created. The controller plugins allows users to push jobs by its service name.

```php
namespace MyModule\Controller;

use Zend\Mvc\Controller\AbstractActionController;

class MyController extends AbstractActionController
{
    public function fooAction()
    {
        // Do some work

        $this->queue('default')
             ->push('MyModule\Job\SendEmailJob', array(
                 'to'      => 'bob@acme.com',
                 'subject' => 'Registration completed',
                 'message' => 'Hi bob, you just registered for our website! Welcome!'
             ));
    }
}
```

The second parameter from the plugin's `push()` is optional. You can omit the array completely. Also, it is possible to
inject multiple jobs into the same queue, or inject a job in another queue:

```php
namespace MyModule\Controller;

use Zend\Mvc\Controller\AbstractActionController;

class MyController extends AbstractActionController
{
    public function fooAction()
    {
        // Do some work

        // Push two jobs into the same queue
        $this->queue('default')
             ->push('MyModule\Job\SendEmailJob', array(
                 'to'      => 'bob@acme.com',
                 'subject' => 'Registration completed',
                 'message' => 'Hi bob, you just registered for our website! Welcome!'

             ))
             ->push('MyModule\Job\TweetJob', array(
                 'message' => 'We have a new user registered, his name is bob!'
             ));

        // Push into anther queue
        $this->queue('background')
             ->push('OtherModule\Job\BussinessReportJob');
    }
}
```

Job status codes
----------------

When using [events](6.Events.md) you might want to hook in the status process of a job. Has
a job successfully been executed or were there errors? The result of a job is expressed in
its status code. SlmQueue defines the following default status codes:

0. `JOB_STATUS_UNKNOWN`
1. `JOB_STATUS_SUCCESS`
2. `JOB_STATUS_FAILURE`
3. `JOB_STATUS_FAILURE_RECOVERABLE`

The status codes are stored in the WorkerEvent object (more on that at the 
[event section](6.Events.md)). Normally when jobs are completely executed, the status is 
success. If any exception is thrown, the status is set to failure.

```php
use SlmQueue\Job\AbstractJob;

class SuccessfulJob extends AbstractJob
{
    public function execute()
    {
        // all is OK
    }
}
```

```php
use RuntimeException
use SlmQueue\Job\AbstractJob;

class FailingJob extends AbstractJob
{
    public function execute()
    {
        throw new RuntimeException('Not going well');
    }
}
```

However, if you want to indicate `JOB_STATUS_FAILURE_RECOVERABLE` or you need to introduce
some custom job status codes, you can return a non-NULL value from the Job's `execute()`
method:

```php
use SlmQueue\Job\AbstractJob;
use SlmQueue\Worker\WorkerEvent;

class RecoverableJob extends AbstractJob
{
    public function execute()
    {
        // Ooops, something went wrong?
        return WorkerEvent::JOB_STATUS_FAILURE_RECOVERABLE;
    }
}
```

Navigation
----------

Previous page: [Configuration](2.Configuration.md)
Next page: [Queue Aware](4.QueueAware.md)

1. [Introduction](1.Introduction.md)
2. [Configuration](2.Configuration.md)
3. [Jobs](3.Jobs.md)
4. [QueueAware](4.QueueAware.md)
5. [Workers](5.Workers.md)
6. [Events](6.Events.md)
7. [Worker management](7.WorkerManagement.md)
